import zlib
import os

from .headers import headers_of_filename


def extract_file_based_on_header_info(apk_file, offset, header_info):
    """
    Extracts a single file from the apk_file based on the information provided from the offset and the header_info.
    It takes into account that the compression method provided might not be STORED or DEFLATED and in that case
    it treats it as STORED -> TODO: inspect other cases
    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param offset: The offset at which the local header for that file is.
    :param header_info: The local header dictionary info for that specific filename (parse_local_header)
    :return: Returns the actual extracted data for that file
    """
    filename_length = header_info["File name length"]
    extra_field_length = header_info["Extra field length"]
    compressed_size = header_info["Compressed size"]
    uncompressed_size = header_info["Uncompressed size"]
    compression_method = header_info["Compression method"]
    # Skip the offset + local header to reach the compressed data
    local_header_size = 30  # Size of the local header in bytes
    apk_file.seek(offset + local_header_size + filename_length + extra_field_length)
    if compression_method == 0:  # Stored (no compression)
        compressed_data = apk_file.read(compressed_size)
        extracted_data = compressed_data
    elif compression_method == 8:
        compressed_data = apk_file.read(compressed_size)
        # -15 for windows size due to raw stream with no header or trailer
        extracted_data = zlib.decompress(compressed_data, -15)
    else:
        # Any ZIP compression method other than STORED is treated as DEFLATED by Android.
        try:
            cur_loc = apk_file.tell()
            compressed_data = apk_file.read(compressed_size)
            extracted_data = zlib.decompress(compressed_data, -15)
        except:
            apk_file.seek(cur_loc)
            compressed_data = apk_file.read(uncompressed_size)
            extracted_data = compressed_data
    return extracted_data


def extract_all_files_from_central_directory(apk_file, central_directory_entries, output_dir):
    """
    Extracts all files from an APK based on the entries detected in the central_directory_entries.
    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param central_directory_entries: The dictionary with all the entries for the central directory (see parse_central_directory)
    :param output_dir: The output directory where to save the files.
    :return: Returns 0 if no errors, 1 if an exception and 2 if the output directory already exists
    """
    try:
        # Check if the output directory already exists
        if os.path.exists(output_dir):
            print("Extraction aborted. Output directory already exists.")
            return 2
        # Create the output directory or overwrite if it already exists
        os.makedirs(output_dir, exist_ok=True)
        # Iterate over central directory entries
        for filename, cd_header_info in central_directory_entries.items():
            # Get the local header offset from the central directory entry
            local_header_offset = cd_header_info["Relative offset of local file header"]
            # Retrieve the header information for the file
            _, header_info = headers_of_filename(apk_file, central_directory_entries, filename)
            # Extract the file using the local header information
            extracted_data = extract_file_based_on_header_info(apk_file, local_header_offset, header_info)
            # Construct the output file path
            output_path = os.path.join(output_dir, filename)
            # Create directories if necessary
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            # Write the extracted data to the output file
            with open(output_path, 'wb') as output_file:
                output_file.write(extracted_data)
        return 0
    except Exception as e:
        print(f"Error extracting files: {e}")
        return 1



import argparse
import io
import os

from apkInspector import __version__ as lib_version
from . import __version__ as cli_version
from apkInspector.extract import extract_file_based_on_header_info, extract_all_files_from_central_directory
from apkInspector.headers import find_eocd, parse_central_directory, headers_of_filename, print_headers_of_filename, \
    get_and_save_local_headers_of_all, show_and_save_info_of_central
from apkInspector.helpers import save_data_to_file
from apkInspector.manifestDecoder import ResChunkHeader, StringPoolType, process_headers, get_manifest


def main():
    parser = argparse.ArgumentParser(description='apkInspector is a tool designed to provide detailed insights into '
                                                 'the central directory and local headers of APK files, offering the '
                                                 'capability to extract content and decode the AndroidManifest.xml '
                                                 'file.')
    parser.add_argument('-apk', help='APK to inspect')
    parser.add_argument('-f', '--filename', help='Filename to provide info for')
    parser.add_argument('-ll', '--list-local', action='store_true', help='List all files by name from local headers')
    parser.add_argument('-lc', '--list-central', action='store_true', help='List all files by name from central '
                                                                           'directory header')
    parser.add_argument('-la', '--list-all', action='store_true',
                        help='List all files from both central directory and local headers')
    parser.add_argument('-e', '--export', action='store_true',
                        help='Export to JSON. What you list from the other flags, will be exported')
    parser.add_argument('-x', '--extract', action='store_true', help='Attempt to extract the file specified by the -f '
                                                                     'flag')
    parser.add_argument('-xa', '--extract-all', action='store_true', help='Attempt to extract all files detected in '
                                                                          'the central directory header')
    parser.add_argument('-m', '--manifest', action='store_true',
                        help='Extract and decode the AndroidManifest.xml')
    parser.add_argument('-v', '--version', action='store_true', help='Retrieves version information')
    args = parser.parse_args()

    if args.version:
        mm = """
#                 _     _____                                 _                
#                | |   |_   _|                               | |               
#    __ _  _ __  | | __  | |   _ __   ___  _ __    ___   ___ | |_   ___   _ __ 
#   / _` || '_ \ | |/ /  | |  | '_ \ / __|| '_ \  / _ \ / __|| __| / _ \ | '__|
#  | (_| || |_) ||   <  _| |_ | | | |\__ \| |_) ||  __/| (__ | |_ | (_) || |   
#   \__,_|| .__/ |_|\_\ \___/ |_| |_||___/| .__/  \___| \___| \__| \___/ |_|   
#         | |                             | |                                  
#         |_|                             |_|                                  
"""
        print(mm)
        print(f"apkInspector CLI Version: {cli_version}")
        print(f"apkInspector Library Version: {lib_version}")
        print(f"Copyright 2023 erev0s <projects@erev0s.com>\n")
        return
    print(f"apkInspector CLI Version: {cli_version}")
    print(f"apkInspector Library Version: {lib_version}")
    print(f"Copyright 2023 erev0s <projects@erev0s.com>\n")
    if args.apk is None:
        print("APK file is required")
        return
    apk_name = os.path.splitext(args.apk)[0]
    with open(args.apk, 'rb') as apk_file:
        eocd = find_eocd(apk_file)
        if eocd is None:
            print("Are you sure you are trying to parse an APK file?")
            return
        central_directory_entries = parse_central_directory(apk_file, eocd["Offset of start of central directory"])

        if args.filename and args.extract:
            try:
                cd_h_of_file, local_header_of_file = headers_of_filename(apk_file, central_directory_entries, args.filename)
            except TypeError as e:
                print(f"Are you sure the filename: {args.filename} exists?")
                exit()
            offset = cd_h_of_file["Relative offset of local file header"]
            print_headers_of_filename(cd_h_of_file, local_header_of_file)
            extracted_data = extract_file_based_on_header_info(apk_file, offset, local_header_of_file)
            save_data_to_file(f"EXTRACTED_{args.filename}", extracted_data)
        elif args.filename:
            try:
                cd_h_of_file, local_header_of_file = headers_of_filename(apk_file, central_directory_entries, args.filename)
            except TypeError as e:
                print(f"Are you sure the filename: {args.filename} exists?")
                exit()
            print_headers_of_filename(cd_h_of_file, local_header_of_file)
        elif args.extract_all:
            print(f"Number of entries: {len(central_directory_entries)}")
            if not extract_all_files_from_central_directory(apk_file, central_directory_entries, apk_name):
                print(f"Extraction successful for: {apk_name}")
        elif args.list_local:
            get_and_save_local_headers_of_all(apk_file, central_directory_entries, apk_name, args.export)
            print(f"Local headers list complete. Export: {args.export}")
        elif args.list_central:
            show_and_save_info_of_central(central_directory_entries, apk_name, args.export)
            print(f"Central header list complete. Export: {args.export}")
        elif args.list_all:
            show_and_save_info_of_central(central_directory_entries, apk_name, args.export)
            get_and_save_local_headers_of_all(apk_file, central_directory_entries, apk_name, args.export)
            print(f"Central and local headers list complete. Export: {args.export}")
        elif args.manifest:
            cd_h_of_file, local_header_of_file = headers_of_filename(apk_file, central_directory_entries,
                                                                     "AndroidManifest.xml")
            offset = cd_h_of_file["Relative offset of local file header"]
            extracted_data = io.BytesIO(extract_file_based_on_header_info(apk_file, offset, local_header_of_file))
            ResChunkHeader.from_file(extracted_data)
            string_pool = StringPoolType.from_file(extracted_data)
            string_data = string_pool.strdata
            elements = process_headers(extracted_data)
            manifest = get_manifest(elements, string_data)
            with open("decoded_AndroidManifest.xml", "w", encoding="utf-8") as xml_file:
                xml_file.write(manifest)
            print("AndroidManifest was saved as: decoded_AndroidManifest.xml")
        else:
            parser.print_help()


if __name__ == '__main__':
    main()



import struct


class ResChunkHeader:
    """
    It is the starting header of the AndroidManifest.xml file.
    Although the header specification states to start with \x03\x00 it is not validate by Android itself.
    """

    def __init__(self, header_type, header_size, total_size):
        self.type = header_type
        self.header_size = header_size
        self.total_size = total_size

    @classmethod
    def from_file(cls, file):
        """
        Read the header type (2 bytes), header size (2 bytes), and entry size (4 bytes)
        :param file: the xml file e.g. with open('/path/AndroidManifest.xml', 'rb') as file:
        :return: Returns an instance of itself
        """
        header_data = file.read(8)
        if len(header_data) < 8:
            # End of file
            return None
        header_type, header_size, total_size = struct.unpack('<HHI', header_data)
        return cls(header_type, header_size, total_size)


class ResStringPoolHeader:
    """
    It reads the string pool header which contains information about the StringCount etc.
    It is a common practice to have a stringCount value that does not reflect the actual strings in place. Android
    itself does not take into account the stringCount when getting the string offsets and the string data!
    """

    def __init__(self, header_type, header_size, total_size, string_count, style_count, flags, strings_start,
                 styles_start):
        self.header = ResChunkHeader(header_type, header_size, total_size)
        self.string_count = string_count
        self.style_count = style_count
        self.flags = flags
        self.strings_start = strings_start
        self.styles_start = styles_start

    @classmethod
    def from_file(cls, file):
        """
        Read and parse ResStringPoolHeader from the file
        :param file: the xml file right after the header has been read.
        :return: Returns an instance of itself
        """
        header = ResChunkHeader.from_file(file)
        string_pool_header_data = file.read(20)
        string_count, style_count, flags, strings_start, styles_start = struct.unpack('<IIIII', string_pool_header_data)
        return cls(header.type, header.header_size, header.total_size, string_count, style_count, flags, strings_start,
                   styles_start)


class StringPoolType:
    """
    The stringPool class which contains the header defined before: ResStringPoolHeader
    along with the string offsets and the string data.
    """
    def __init__(self, header_type, header_size, total_size, string_count, style_count, flags, strings_start,
                 styles_start, string_offsets, strdata):
        self.header = ResStringPoolHeader(header_type, header_size, total_size, string_count, style_count, flags,
                                          strings_start,
                                          styles_start)
        self.string_offsets = string_offsets
        self.strdata = strdata

    @classmethod
    def read_string_offsets(cls, file, num_of_strings, end_absolute_offset):
        """
        Reads the offsets available for each string. The important thing to notice is that the number of strings being
        passed as a parameter should not be the stringCount retrieved from the string pool header, it should be
        calculated!
        :param file: the xml file right after the string pool header has been read.
        :param num_of_strings: the calculated number of strings available
        :param end_absolute_offset: the absolute value of the offset where the offsets finish.
        :return: Returns a list of strings offsets.
        """
        string_offsets = []
        for i in range(0, num_of_strings):
            string_offsets.append(struct.unpack('<I', file.read(4))[0])
        # sanity check as after reading the last string we should be at the end offset as calculated
        if file.tell() != end_absolute_offset:
            print(
                f"Current file read:{file.tell()} is not as expected to the start of the stringData:{end_absolute_offset})")
        return string_offsets

    @classmethod
    def decode_stringpool_mixed_string(cls, file, is_utf8):
        """
        Handling the different encoding possibilities that can be met.
        :param file: the xml file at the offset where the string is to be read
        :param is_utf8: boolean to check if a utf8 string is expected
        :return: Returns the decoded string
        """
        if not is_utf8:
            # Handle UTF-16 encoded strings
            u16len = struct.unpack('<H', file.read(2))[0]
            if u16len & 0x8000 == 0:
                # Regular UTF-16 string
                content = file.read(u16len * 2).decode('utf-16le')
            else:
                # UTF-16 string with fixup
                u16len_fix = struct.unpack('<H', file.read(2))[0]
                real_length = ((u16len & 0x7FFF) << 16) | u16len_fix
                content = file.read(real_length * 2).decode('utf-16le')
        else:
            # Handle UTF-8 encoded strings
            u16len = struct.unpack('B', file.read(1))[0]
            file.read(1)
            u8len = u16len
            content = file.read(u8len).decode('utf-8', errors='replace')

        return content

    @classmethod
    def read_strings(cls, file, string_offsets, strings_start, is_utf8):
        """
        Gets the actual strings based on the offsets retrieved from read_string_offsets()
        :param file: the xml file right after the string pool offsets have been read
        :param string_offsets: see -> read_string_offsets()
        :param strings_start:
        :param is_utf8: the offset at which the string data starts
        :return: Returns a list of the string data
        """
        strings = []
        for offset in string_offsets:
            # Calculate the absolute offset within the string data +8 for the file header
            absolute_offset = strings_start + offset + 8  # TODO: update this to get the file header size
            # Move the file pointer to the start of the string
            file.seek(absolute_offset)
            # Read the length of the string (in bytes)
            content = cls.decode_stringpool_mixed_string(file, is_utf8)
            strings.append(content)
        return strings

    @classmethod
    def from_file(cls, file):
        """
        Handle the string pool
        :param file: the xml file right after the file header is read
        :return: Returns an instance of itself
        """
        string_pool_header = ResStringPoolHeader.from_file(file)
        size_of_strings_offsets = string_pool_header.strings_start - 28
        # it should be divisible by 4, as 4 bytes are per offset, so we can get accurately the # of strings
        num_of_strings = size_of_strings_offsets // 4
        if not (size_of_strings_offsets / 4).is_integer():
            print(f"The number of strings in the string pool is not a integer number.")
        string_offsets = cls.read_string_offsets(file, num_of_strings, string_pool_header.strings_start + 8)
        is_utf8 = bool(string_pool_header.flags & (1 << 8))
        string_data = cls.read_strings(file, string_offsets, string_pool_header.strings_start, is_utf8)
        file.read(2)  # the +2 is to account for the null bytes after the last strPool element
        return cls(
            string_pool_header.header.type,
            string_pool_header.header.header_size,
            string_pool_header.header.total_size,
            string_pool_header.string_count,
            string_pool_header.style_count,
            string_pool_header.flags,
            string_pool_header.strings_start,
            string_pool_header.styles_start,
            string_offsets,
            string_data
        )


class XmlResourceMapType(ResChunkHeader):
    """
    Resource map class, with the header and the resource IDs
    """
    def __init__(self, header_type, header_size, total_size, resids):
        super().__init__(header_type, header_size, total_size)
        self.resids = resids

    @classmethod
    def parse(cls, file, header: ResChunkHeader):
        """
        Parse the resourse map and get the resource IDs
        :param file: the xml file right after the string pool is read
        :param header: the header ResChunkHeader
        :return: Returns an instance of itself
        """
        num_resids = (header.total_size - header.header_size) // 4
        resids = [struct.unpack('<I', file.read(4))[0] for _ in range(num_resids)]

        return cls(header.type, header.header_size, header.total_size, resids)


class XmlStartNamespace(ResChunkHeader):
    """
    The actual start of the xml, after this the elements of the xml will be found.
    """
    def __init__(self, header_type, header_size, total_size, ext):
        super().__init__(header_type, header_size, total_size)
        self.ext = ext

    @classmethod
    def parse(cls, file, header: ResChunkHeader):
        num_exts = (header.total_size - header.header_size) // 4
        ext = [struct.unpack('<I', file.read(4))[0] for _ in range(num_exts)]

        return cls(header.type, header.header_size, header.total_size, ext)


class XmlEndNamespace(ResChunkHeader):
    """
    Indicator for the end of the xml file.
    """
    def __init__(self, header_type, header_size, total_size, prefix_namespace_index, uri_index):
        super().__init__(header_type, header_size, total_size)
        self.prefix_namespace_index = prefix_namespace_index
        self.uri_index = uri_index

    @classmethod
    def parse(cls, file, header: ResChunkHeader):
        prefix_namespace_index = struct.unpack('<I', file.read(4))[0]
        uri_index = struct.unpack('<I', file.read(4))[0]

        return cls(header.type, header.header_size, header.total_size, prefix_namespace_index, uri_index)


class XmlAttributeElement:
    """
    The attributes within each element within the xml, should be described by this class
    """
    def __init__(self, full_namespace_index, name_index, raw_value_index, typed_value_size, typed_value_res0,
                 typed_value_datatype, typed_value_data):
        self.full_namespace_index = full_namespace_index
        self.name_index = name_index
        self.raw_value_index = raw_value_index
        self.typed_value_size = typed_value_size
        self.typed_value_res0 = typed_value_res0
        self.typed_value_datatype = typed_value_datatype
        self.typed_value_data = typed_value_data

    @classmethod
    def parse(cls, file, attr_count, attr_size):
        attrs = []
        for _ in range(0, attr_count):
            full_namespace_index = struct.unpack('<I', file.read(4))[0]
            name_index = struct.unpack('<I', file.read(4))[0]
            raw_value_index = struct.unpack('<I', file.read(4))[0]
            typed_value_size = struct.unpack('<H', file.read(2))[0]
            typed_value_res0 = struct.unpack('<B', file.read(1))[0]
            typed_value_datatype = struct.unpack('<B', file.read(1))[0]
            typed_value_data = struct.unpack('<I', file.read(4))[0]
            attrs.append(cls(full_namespace_index, name_index, raw_value_index, typed_value_size, typed_value_res0,
                             typed_value_datatype, typed_value_data))
        return attrs


class XmlStartElement(ResChunkHeader):
    """
    The starting point of an element, its attributes are described by XmlAttributeElement
    The attrext contains information about the element including the attribute count.
    """
    def __init__(self, header_type, header_size, total_size, attrext, attributes):
        super().__init__(header_type, header_size, total_size)
        self.attrext = attrext
        self.attributes = attributes

    @classmethod
    def parse(cls, file, header: ResChunkHeader):
        full_namespace_index, name_index, attr_start, attr_size, attr_count, id_index, class_index, style_index = struct.unpack(
            '<IIHHHHHH', file.read(20))

        attrext = [full_namespace_index, name_index, attr_start, attr_size, attr_count, id_index, class_index,
                   style_index]
        attributes = XmlAttributeElement.parse(file, attr_count, attr_size)
        return cls(header.type, header.header_size, header.total_size, attrext, attributes)


class XmlEndElement(ResChunkHeader):
    """
    The end of an element, where the attrext contains the necessary information on which element it ends.
    """
    def __init__(self, header_type, header_size, total_size, attrext):
        super().__init__(header_type, header_size, total_size)
        self.attrext = attrext

    @classmethod
    def parse(cls, file, header: ResChunkHeader):
        full_namespace_index, name_index = struct.unpack('<II', file.read(8))
        attrext = [full_namespace_index, name_index]
        return cls(header.type, header.header_size, header.total_size, attrext)


def process_xml_resource_map(file, chunk_header: ResChunkHeader):
    return XmlResourceMapType.parse(file, chunk_header)


def process_xml_start_namespace(file, chunk_header: ResChunkHeader):
    return XmlStartNamespace.parse(file, chunk_header)


def process_xml_end_namespace(file, chunk_header: ResChunkHeader):
    return XmlEndNamespace.parse(file, chunk_header)


def process_xml_start_element(file, chunk_header: ResChunkHeader):
    return XmlStartElement.parse(file, chunk_header)


def process_xml_end_element(file, chunk_header: ResChunkHeader):
    return XmlEndElement.parse(file, chunk_header)


chunk_type_handlers = {
    '0x180': process_xml_resource_map,      # RES_XML_RESOURCE_MAP_TYPE
    '0x100': process_xml_start_namespace,   # RES_XML_START_NAMESPACE_TYPE
    '0x102': process_xml_start_element,     # RES_XML_START_ELEMENT_TYPE
    '0x103': process_xml_end_element,       # RES_XML_END_ELEMENT_TYPE
    '0x101': process_xml_end_namespace,      # RES_XML_END_NAMESPACE_TYPE
}


def parse_next_header(file):
    """
    Dispatcher method to parse the next available header. It takes into account to move on past the header if it
    contains extra info besides the standard ones.
    The dispatcher automatically picks the correct processing method for each chunk type.
    :param file: the xml file that is being read
    :return: Dispatches to the appropriate processing method for each chunk type.
    """
    chunk_header = ResChunkHeader.from_file(file)
    if chunk_header is None:  # end of file
        return None
    if chunk_header.header_size > 8:
        # read the rest of the header
        file.read(chunk_header.header_size - 8)
    chunk_type = hex(chunk_header.type)
    if chunk_type in chunk_type_handlers:
        return chunk_type_handlers[chunk_type](file, chunk_header)
    else:
        raise NotImplementedError(f"Unsupported chunk type: {chunk_type}")


def process_headers(file):
    """
    It starts processing the remaining headers after the string pool chunk.
    Takes into account that the resource map, the start namespace and end namespace chunks are only to be found once
    within the file.
    :param file: the xml file after the string pool chunk
    :return: Returns all the elements found as their corresponding classes
    """
    elements = []
    possible_headers = {b'\x80\x01', b'\x00\x01', b'\x02\x01', b'\x03\x01', b'\x01\x01'}
    while True:
        # Parse the next header
        cur_pos = file.tell()
        check = file.read(2)
        file.seek(cur_pos)
        if not check:
            # End of file
            break
        if check not in possible_headers:
            file.read(1)
            continue
        chunk_type = parse_next_header(file)
        elements.append(chunk_type)
        if check in {b'\x80\x01', b'\x00\x01', b'\x01\x01'}:
            possible_headers.remove(check)
    return elements


def get_manifest(elements, string_data):
    """
    Method to go over all elements and attempt to create the AndroidManifest.xml file.
    :param elements: Elements as retrieved from process_headers()
    :param string_data: The string data contained within the string pool
    :return: Returns the string of the AndroidManifest.xml
    """
    android_manifest_xml = '<?xml version="1.0" encoding="utf-8"?>\n'
    namespace = ""
    for element in elements:
        if isinstance(element, XmlStartNamespace):
            namespace = f'xmlns:{string_data[element.ext[0]]}="{string_data[element.ext[1]]}"'
        elif isinstance(element, XmlStartElement):
            ln_list = []
            for attr in element.attributes:
                if attr.typed_value_datatype == 1:  # reference type
                    name = string_data[attr.name_index]
                    value = f"@{attr.typed_value_data}"
                elif attr.typed_value_datatype == 3:  # string type
                    name = string_data[attr.name_index]
                    try:
                        value = string_data[attr.typed_value_data]
                    except:
                        value = attr.typed_value_data
                elif attr.typed_value_datatype == 16:  # integer type
                    name = string_data[attr.name_index]
                    value = attr.typed_value_data
                elif attr.typed_value_datatype == 17:  # int-hex type
                    name = string_data[attr.name_index]
                    value = f"{attr.typed_value_data} ({hex(attr.typed_value_data)})"
                elif attr.typed_value_datatype == 18:  # boolean type
                    name = string_data[attr.name_index]
                    value = "true" if bool(attr.typed_value_data) else "false"
                else:
                    print(f"An unknown datatype came up: {attr.typed_value_datatype}")
                ln_list.append(f'android:{name}="{value}"')
            if ln_list:
                if string_data[element.attrext[1]] == "manifest":
                    android_manifest_xml += f"<{string_data[element.attrext[1]]} {namespace} {' '.join(ln_list)}>\n"
                else:
                    android_manifest_xml += f"<{string_data[element.attrext[1]]} {' '.join(ln_list)}>\n"
            else:
                android_manifest_xml += f"<{string_data[element.attrext[1]]}>\n"
        elif isinstance(element, XmlEndElement):
            name = string_data[element.attrext[1]]
            if name == "manifest":
                android_manifest_xml += f"</{string_data[element.attrext[1]]}>"
            else:
                android_manifest_xml += f"</{string_data[element.attrext[1]]}>\n"
    return android_manifest_xml


# def beautify_validate_xml(xml_string):
#     import xml.etree.ElementTree as ET
#     ET.register_namespace("android", "http://schemas.android.com/apk/res/android")
#     root = ET.fromstring(xml_string)
#     xml_pretty = ET.tostring(root, encoding='utf-8', method='xml').decode("utf-8")
#     return xml_pretty






import struct

from .helpers import pretty_print_header, save_to_json


def find_eocd(apk_file):
    """
    Method to locate the "end of central directory record signature" as the first step of the correct process of
    reading a ZIP archive. Should be noted that certain APKs do not follow the zip specification and declare multiple
    "end of central directory records". For this reason the search for the corresponding signature of the eocd starts
    from the end of the apk.
    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :return: Returns the end of central directory record with all the information available if the corresponding
    signature is found. If not, then it returns None.
    """
    chunk_size = 1024
    offset = 0
    file_size = apk_file.seek(0, 2)
    while offset < file_size:
        position = file_size - offset - chunk_size
        if position < 0:
            position = 0
        apk_file.seek(position)
        chunk = apk_file.read(chunk_size)
        if not chunk:
            break
        signature_offset = chunk.rfind(b'\x50\x4b\x05\x06')  # end of Central Directory File Header signature
        if signature_offset != -1:
            eo_central_directory_offset = position + signature_offset
            break  # Found End of central directory record (EOCD) signature
        offset += chunk_size
    if signature_offset == -1:
        return None  # End of central directory record (EOCD) signature not found
    apk_file.seek(eo_central_directory_offset)
    eocd = {}
    signature = apk_file.read(4)
    eocd["Number of this disk"] = struct.unpack('<H', apk_file.read(2))[0]
    eocd["Disk where central directory starts"] = struct.unpack('<H', apk_file.read(2))[0]
    eocd["Number of central directory records on this disk"] = struct.unpack('<H', apk_file.read(2))[0]
    eocd["Total number of central directory records"] = struct.unpack('<H', apk_file.read(2))[0]
    eocd["Size of central directory (bytes)"] = struct.unpack('<I', apk_file.read(4))[0]
    eocd["Offset of start of central directory"] = struct.unpack('<I', apk_file.read(4))[0]
    eocd["Comment length"] = struct.unpack('<H', apk_file.read(2))[0]
    eocd["Comment"] = struct.unpack(f'<{eocd["Comment length"]}s', apk_file.read(eocd["Comment length"]))[0]

    return eocd


def parse_central_directory(apk_file, central_directory_offset):
    """
    Method that is used to parse the central directory header according to the specification
    https://pkware.cachefly.net/webdocs/APPNOTE/APPNOTE-6.3.9.TXT
    based on the offset provided by the end of central directory record: eocd["Offset of start of central directory"].
    If multiple central directory headers are discovered this will not be handled properly!

    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param central_directory_offset: The offset as read from the end of central
    directory record : eocd["Offset of start of central directory"]
    :return: Returns a dictionary with all the entries discovered. The filename of each entry is used as the key. Besides
    the fields defined by the specification, each entry has an additional field named 'Offset in the central directory header',
    which includes the offset of the entry in the central directory itself.
    """

    apk_file.seek(central_directory_offset)
    if apk_file.tell() != central_directory_offset:
        raise ValueError(f"Failed to find the offset for the central directory within the file!")

    central_directory_entries = {}
    while True:
        c_offset = apk_file.tell()
        signature = apk_file.read(4)
        if signature != b'\x50\x4b\x01\x02':
            break  # Reached the end of the central directory
        central_directory_entry = {"Version made by": struct.unpack('<H', apk_file.read(2))[0],
                                   "Version needed to extract (minimum)": struct.unpack('<H', apk_file.read(2))[0],
                                   "General purpose bit flag": struct.unpack('<H', apk_file.read(2))[0],
                                   "Compression method": struct.unpack('<H', apk_file.read(2))[0],
                                   "File last modification time": struct.unpack('<H', apk_file.read(2))[0],
                                   "File last modification date": struct.unpack('<H', apk_file.read(2))[0],
                                   "CRC-32 of uncompressed data": struct.unpack('<I', apk_file.read(4))[0],
                                   "Compressed size": struct.unpack('<I', apk_file.read(4))[0],
                                   "Uncompressed size": struct.unpack('<I', apk_file.read(4))[0],
                                   "File name length": struct.unpack('<H', apk_file.read(2))[0],
                                   "Extra field length": struct.unpack('<H', apk_file.read(2))[0],
                                   "File comment length": struct.unpack('<H', apk_file.read(2))[0],
                                   "Disk number where file starts": struct.unpack('<H', apk_file.read(2))[0],
                                   "Internal file attributes": struct.unpack('<H', apk_file.read(2))[0],
                                   "External file attributes": struct.unpack('<I', apk_file.read(4))[0],
                                   "Relative offset of local file header": struct.unpack('<I', apk_file.read(4))[0]}
        filename_length = central_directory_entry["File name length"]
        central_directory_entry["Filename"] = struct.unpack(f'<{filename_length}s', apk_file.read(filename_length))[
            0].decode('utf-8')
        extra_field_length = central_directory_entry["Extra field length"]
        central_directory_entry["Extra Field"] = \
            struct.unpack(f'<{extra_field_length}s', apk_file.read(extra_field_length))[0].decode('utf-8', 'ignore')
        file_comment_length = central_directory_entry["File comment length"]
        central_directory_entry["File comment"] = \
            struct.unpack(f'<{file_comment_length}s', apk_file.read(file_comment_length))[0].decode('utf-8', 'ignore')

        central_directory_entry["Offset in the central directory header"] = c_offset
        central_directory_entries[central_directory_entry["Filename"]] = central_directory_entry

    return central_directory_entries


def parse_local_header(apk_file, entry_of_interest):
    """
    Method that attempts to read the local file header according to the specification
    https://pkware.cachefly.net/webdocs/APPNOTE/APPNOTE-6.3.9.TXT

    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param entry_of_interest: The central directory header of the specific entry of interest
    :return: Returns a dictionary with the local header information or None if it failed to find the header.
    """
    apk_file.seek(entry_of_interest["Relative offset of local file header"])
    header_signature = apk_file.read(4)

    if not header_signature == b'\x50\x4b\x03\x04':
        print(f"Does not seem to be the start of a local header!")
        return None
    else:
        local_header_info = {"Version needed to extract (minimum)": struct.unpack('<H', apk_file.read(2))[0],
                             "General purpose bit flag": struct.unpack('<H', apk_file.read(2))[0],
                             "Compression method": struct.unpack('<H', apk_file.read(2))[0],
                             "File last modification time": struct.unpack('<H', apk_file.read(2))[0],
                             "File last modification date": struct.unpack('<H', apk_file.read(2))[0],
                             "CRC-32 of uncompressed data": struct.unpack('<I', apk_file.read(4))[0],
                             "Compressed size": struct.unpack('<I', apk_file.read(4))[0],
                             "Uncompressed size": struct.unpack('<I', apk_file.read(4))[0],
                             "File name length": struct.unpack('<H', apk_file.read(2))[0],
                             "Extra field length": struct.unpack('<H', apk_file.read(2))[0]}

        filename_length = local_header_info["File name length"]
        local_header_info["Filename"] = struct.unpack(f'<{filename_length}s', apk_file.read(filename_length))[0].decode(
            'utf-8')
        extra_field_length = local_header_info["Extra field length"]
        local_header_info["Extra Field"] = struct.unpack(f'<{extra_field_length}s', apk_file.read(extra_field_length))[
            0].decode('utf-8', 'ignore')

    return local_header_info


def headers_of_filename(apk_file, central_directory_entries, filename):
    """
    Provides both the central directory header and the local header of a specific filename
    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param central_directory_entries: The dictionary with all the entries for the central directory (see
    parse_central_directory)
    :param filename: The filename of the entry of interest
    :return: Returns two dictionaries with the central directory header and the local header of the specified filename.
    If the filename is not found within the central directory dictionary it returns None.
    """
    if filename in central_directory_entries:
        cd_h_of_file = central_directory_entries[filename]
        local_header_of_file = parse_local_header(apk_file, cd_h_of_file)
        return cd_h_of_file, local_header_of_file
    else:
        print(f"Filename: {filename} does not seem to exist within the central directory entries.")
        return None


def print_headers_of_filename(cd_h_of_file, local_header_of_file):
    """
    Prints out the details for both the central directory header and the local file header. Useful for the CLI
    :param cd_h_of_file: central directory header of a filename as it may be retrieved from headers_of_filename
    :param local_header_of_file: local header dictionary of a filename as it may be retrieved from headers_of_filename
    """
    pretty_print_header("CENTRAL DIRECTORY")
    for k in cd_h_of_file:
        if k == 'Relative offset of local file header' or k == 'Offset in the central directory header':
            print(f"{k:40} : {hex(int(cd_h_of_file[k]))} | {cd_h_of_file[k]}")
        else:
            print(f"{k:40} : {cd_h_of_file[k]}")
    pretty_print_header("LOCAL HEADER")
    for k in local_header_of_file:
        print(f"{k:40} : {local_header_of_file[k]}")


def show_and_save_info_of_central(central_directory_entries, apk_name, export: bool):
    """
    Print information for each entry for the central directory header and allow to possibly export to JSON
    :param central_directory_entries: The dictionary with all the entries for the central directory (see parse_central_directory)
    :param apk_name: String with the name of the APK, so it can be used for the export.
    :param export: Boolean for exporting or not to JSON
    """
    if not export:
        for entry in central_directory_entries:
            pretty_print_header(entry)
            print(central_directory_entries[entry])
    else:
        save_to_json(f"{apk_name}_central_directory_header.json", central_directory_entries)


def get_and_save_local_headers_of_all(apk_file, central_directory_entries, apk_name, export: bool):
    """
    Creates a dictionary of local headers based on the entries retrieved from the central directory header.
    Additionally, allows to print the local headers and export the dictionary to JSON
    :param apk_file: The already read/loaded data of the APK file e.g. with open('test.apk', 'rb') as apk_file
    :param central_directory_entries: The dictionary with all the entries for the central directory (see parse_central_directory)
    :param apk_name: String with the name of the APK, so it can be used for the export.
    :param export: Boolean for exporting or not to JSON. If exporting it will not print.
    :return: Returns the dictionary created with all the local headers, where the filename is the key.
    """
    local_headers = {}
    for entry in central_directory_entries:
        entry_local_header = parse_local_header(apk_file, central_directory_entries[entry])
        local_headers[entry_local_header['Filename']] = entry_local_header
        if not export:
            pretty_print_header(entry_local_header['Filename'])
            print(entry_local_header)
    if export:
        save_to_json(f"{apk_name}_local_headers.json", local_headers)
    return local_headers
